/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "math.h"

#include "Utils/SPMatrix.h"

#include "Camera.h"

void Camera::setUpProjection(unsigned newWidth, unsigned newHeight, float fov) {
	pixelWidth  = newWidth;
	pixelHeight = newHeight;

	//calculate projection place size from field of view
	planeWidth  = tan(fov/2);
	planeHeight = planeWidth*newHeight/newWidth;
}

void Camera::setLook(const SPVector &newEye, const SPVector &target, const SPVector &up) {
	eye=newEye;
	
	//calculate normalized vector in viewing direction
	SPVector viewDirection= target - newEye;
	viewDirection.Normalize();

	moveFront = viewDirection;

	//Using cross product calculate vectors in directions
	//of "horizontal" and "vertical" axis of projection plane
	dx= viewDirection % up;
	dy= dx % viewDirection;

	moveSide = dx;

	//scale vectors to reflect projection plane size
	dx*=planeWidth;
	dy*=planeHeight;

	//begin in upper left corner
	upperLeft=viewDirection;
	upperLeft+=dx;
	upperLeft+=dy;

	//scale vectors to have size of single step from current ray to next
	dx*=-2.0f/pixelWidth;
	dy*=-2.0f/pixelHeight;

	moveSide.Normalize();
}

void Camera::prepareCuda(CameraCuda &cameraOut) {
	cameraOut.eye.x = eye.x;
	cameraOut.eye.y = eye.y;
	cameraOut.eye.z = eye.z;

	cameraOut.upperLeft.x = upperLeft.x;
	cameraOut.upperLeft.y = upperLeft.y;
	cameraOut.upperLeft.z = upperLeft.z;

	cameraOut.dx.x = dx.x;
	cameraOut.dx.y = dx.y;
	cameraOut.dx.z = dx.z;

	cameraOut.dy.x = dy.x;
	cameraOut.dy.y = dy.y;
	cameraOut.dy.z = dy.z;
};