/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

__device__ float4 operator+ (float4 &a,float4 &b) {
	float4 res = make_float4(a.x + b.x,a.y + b.y,a.z + b.z,a.w + b.w);
	return res;
};

__device__ float4 operator- (float4 &a,float4 &b) {
	float4 res = make_float4(a.x - b.x,a.y - b.y,a.z - b.z,a.w - b.w);
	return res;
};

__device__ void normalize(float4 &a) {
	float size = a.x*a.x + a.y*a.y + a.z*a.z;
	size = rsqrtf(size);
	a.x *= size;
	a.y *= size;
	a.z *= size;
};

__device__ void operator+= (float4 &a,float4 &b) {
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
	a.w += b.w;
};

__device__ void operator-= (float4 &a,float4 &b) {
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
	a.w -= b.w;
};

__device__ void operator-= (float4 &a,float3 &b) {
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
};

__device__ void operator/= (float4 &a,float3 &b) {
	a.x /= b.x;
	a.y /= b.y;
	a.z /= b.z;
};

__device__ void operator*= (float4 &a,float3 &b) {
	a.x *= b.x;
	a.y *= b.y;
	a.z *= b.z;
};

__device__ float4 operator* (float4 &a,float &b) {
	float4 res = make_float4(a.x * b,a.y * b,a.z * b,a.w * b);
	return res;
};

__device__ float4 operator* (float4 &a,unsigned &b) {
	float c = (float)b;
	float4 res = make_float4(a.x * c,a.y * c,a.z * c,a.w * c);
	return res;
};

__device__ float4 operator* (float4 &a,unsigned short &b) {
	float c = (float)b;
	float4 res = make_float4(a.x * c,a.y * c,a.z * c,a.w * c);
	return res;
};

__device__ float4 cross(float4 &a,float4 &b) {
	float4 res;
	res.x = a.y*b.z - a.z*b.y;
	res.y = a.z*b.x - a.x*b.z;
	res.z = a.x*b.y - a.y*b.x;
	res.w = 0.0f;

	return res;
};

__device__ float dot(float4 &a,float4 &b) {
	return a.x*b.x + a.y*b.y + a.z*b.z + a.w*b.w;
};






__device__ float3 operator* (float3 a,float b) {
	float3 res = make_float3(a.x * b,a.y * b,a.z * b);
	return res;
};

__device__ float3 operator* (float3 &a,float3 &b) {
	float3 res = make_float3(a.x * b.x,a.y * b.y,a.z * b.z);
	return res;
};

__device__ void operator*= (float3 &a,float3 &b) {
	a.x *= b.x;
	a.y *= b.y;
	a.z *= b.z;
};

__device__ void operator*= (float3 &a,float b) {
	a.x *= b;
	a.y *= b;
	a.z *= b;
};

__device__ float3 operator* (float3 &a,unsigned short &b) {
	float3 res = make_float3(a.x * b,a.y * b,a.z * b);
	return res;
};

__device__ float3 rcp(float3 &a) {
	float3 res;
	res.x = 1.0f/a.x;
	res.y = 1.0f/a.y;
	res.z = 1.0f/a.z;
	return res;
};

__device__ float3 min(float3 &a,float3 &b) {
	float3 res;
	res.x = min(a.x,b.x);
	res.y = min(a.y,b.y);
	res.z = min(a.z,b.z);
	return res;
};
__device__ float3 max(float3 &a,float3 &b) {
	float3 res;
	res.x = max(a.x,b.x);
	res.y = max(a.y,b.y);
	res.z = max(a.z,b.z);
	return res;
};


__device__ void operator/= (float3 &a,float3 &b) {
	a.x /= b.x;
	a.y /= b.y;
	a.z /= b.z;
};

__device__ void operator/= (float3 &a,float b) {
	a.x /= b;
	a.y /= b;
	a.z /= b;
};


__device__ float3 operator+ (float3 a,float3 b) {
	float3 res = make_float3(a.x + b.x,a.y + b.y,a.z + b.z);
	return res;
};

__device__ void operator+= (float3 &a,float3 b) {
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
};

__device__ float3 operator- (float3 a,float3 b) {
	float3 res = make_float3(a.x - b.x,a.y - b.y,a.z - b.z);
	return res;
};

__device__ void operator-= (float3 &a,float3 &b) {
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
};

__device__ void normalize(float3 &a) {
	float size = a.x*a.x + a.y*a.y + a.z*a.z;
	size = rsqrtf(size);
	a.x *= size;
	a.y *= size;
	a.z *= size;
};

__device__ float3 cross(float3 &a,float3 &b) {
	float3 res;
	res.x = a.y*b.z - a.z*b.y;
	res.y = a.z*b.x - a.x*b.z;
	res.z = a.x*b.y - a.y*b.x;

	return res;
};

__device__ float dot(float3 &a,float3 &b) {
	return a.x*b.x + a.y*b.y + a.z*b.z;
};

__device__ float dotp(float3 &a,float3 &b) {
	return __fmul_rn(a.x,b.x) + __fmul_rn(a.y,b.y) + __fmul_rn(a.z,b.z);
};

__device__ float3 float4to3(float4 a) {
	float3 res;
	res.x = a.x;
	res.y = a.y;
	res.z = a.z;
	return res;
};