#include "StdAfx.h"

#include "MainWindow.h"
#include "TextureDisplay.h"

#include "bitmap/viewOmni.xpm"
#include "bitmap/walk.xpm"
#include "bitmap/X.xpm"
#include "bitmap/Y.xpm"
#include "bitmap/Z.xpm"

static const wxString RESOLUTIONS_TEXT[] = {wxT("512x512"),wxT("640x480"),wxT("800x600"),wxT("1024x768"),wxT("1024x1024")};
Resolution RESOLUTIONS[] = {{512,512},{640,480},{800,600},{1024,768},{1024,1024}};

static const wxString STRUCTURE_NAME[] = {wxT("Grid"),wxT("KdTree"),wxT("BVH")};
static const wxString FORMAT_FILTER = wxT( "All suported files|*.ply;*.tri;*.mdl;*.iv|PLY format(*.ply)|*.ply|Tri files(*.tri2)|*.tri2|Preprocesed internal format(*.mdl)|*.mdl|Open Inventor(*.iv)|*.iv");

enum {
	Menu_File_Open = 200,
	Menu_File_Quit,

	Resolution_Combo = 300,
	Structure_Combo,

	Tool_Omni	= 400,
	Tool_Walk,
	Tool_X,
	Tool_Y,
	Tool_Z
};

BEGIN_EVENT_TABLE(MainWindow, wxFrame)
	EVT_MENU(Menu_File_Quit,		MainWindow::OnQuit)
	EVT_MENU(Menu_File_Open,		MainWindow::OnFileOpen)	
	EVT_COMBOBOX(Resolution_Combo,	MainWindow::OnResolutionChange)
	EVT_COMBOBOX(Structure_Combo,	MainWindow::OnStructureChange)
	EVT_TOOL_RANGE(Tool_Omni,Tool_Walk,MainWindow::onViewMode)
	EVT_TOOL_RANGE(Tool_X,Tool_Z,	MainWindow::onMainAxis)
END_EVENT_TABLE()

MainWindow::MainWindow(const wxString& title) 
		   :wxFrame(NULL, wxID_ANY, title, wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE & ~wxRESIZE_BORDER)	 {
	menubar = new wxMenuBar;
	file	= new wxMenu;

	file->Append(Menu_File_Open, wxT("&Open"));
	file->AppendSeparator();
	file->Append(Menu_File_Quit, wxT("&Quit"));
	
	menubar->Append(file, wxT("&File"));
	SetMenuBar(menubar);

	wxBitmap* toolBarBitmaps[5];	
	toolBarBitmaps[0] = new wxBitmap(viewOmni);
	toolBarBitmaps[1] = new wxBitmap(walkIcon);
	toolBarBitmaps[2] = new wxBitmap(X);
	toolBarBitmaps[3] = new wxBitmap(Y);
	toolBarBitmaps[4] = new wxBitmap(Z);

	CreateToolBar(wxNO_BORDER|wxHORIZONTAL|wxTB_FLAT, -1);
	toolBar = GetToolBar();
	omni = toolBar->AddTool(Tool_Omni,*(toolBarBitmaps[0]),wxNullBitmap,true,0,wxEmptyString);
	walk = toolBar->AddTool(Tool_Walk,*(toolBarBitmaps[1]),wxNullBitmap,true,0,wxEmptyString);
	toolBar->AddSeparator();
	toolBar->AddTool(Tool_X,*(toolBarBitmaps[2]),wxNullBitmap,true);
	toolBar->AddTool(Tool_Y,*(toolBarBitmaps[3]),wxNullBitmap,true);
	toolBar->AddTool(Tool_Z,*(toolBarBitmaps[4]),wxNullBitmap,true);
	toolBar->Realize();

	toolBar->ToggleTool(Tool_Omni,true);
	toolBar->ToggleTool(Tool_X	 ,true);

	axisLast = Tool_X;
	
	delete toolBarBitmaps[0];
	delete toolBarBitmaps[2];
	delete toolBarBitmaps[3];
	delete toolBarBitmaps[4];


	//Create the right panel that holds the info and settings control. 
	settingsPanel = new wxPanel(this);
	//The info panels will be verticly placed.
	wxBoxSizer*	settingsSizer = new wxBoxSizer(wxVERTICAL);
	settingsPanel->SetSizer(settingsSizer);

	//Create the info panels,set their sizes and add them to the sizer.
	wxPanel* structurePanel = new wxPanel(settingsPanel);
	wxPanel* renderingPanel = new wxPanel(settingsPanel);

	structurePanel->SetMinSize(wxSize(180,120));
	renderingPanel->SetMinSize(wxSize(180,90));
	renderingPanel->SetMaxSize(wxSize(180,2000));

	settingsSizer->Add(structurePanel);
	settingsSizer->Add(renderingPanel,1,wxEXPAND);

	//Create the OpenGl canvas.
	canvas = new TextureDisplay(this, wxID_ANY, wxDefaultPosition,wxSize(512, 512), wxSUNKEN_BORDER);

	//Create the main window layout with OpenGl canvas on the left and the info panel on the right.
	wxBoxSizer *mainSizer = new wxBoxSizer(wxHORIZONTAL);
	mainSizer->Add(canvas);
	mainSizer->Add(settingsPanel);

	//CreateStatusBar(1);
	//statusBar = GetStatusBar();
	//statusBar->SetStatusText(wxT("Ready"));
	//SetStatusText(wxT("Ready"));

	//	wxStaticBox* structureBox = new wxStaticBox(settingsPanel, wxID_ANY,wxT("Acceleration structure"));

	//Create controls for the "Rendering" panel.
	new wxStaticBox (renderingPanel, wxID_ANY,wxT("Rendering") ,wxDefaultPosition,wxSize(180,90));
	new wxStaticText(renderingPanel, wxID_ANY,wxT("Resolution"),wxPoint(9,24));
	resolutionBox = new wxComboBox (renderingPanel,Resolution_Combo,wxEmptyString,wxPoint(92,20),wxDefaultSize,5,RESOLUTIONS_TEXT,wxCB_READONLY);
	resolutionBox->SetSelection(0);

	new wxStaticText(renderingPanel, wxID_ANY,wxT("Frame time") ,wxPoint(9,44));
	new wxStaticText(renderingPanel, wxID_ANY,wxT("Frame rate")	,wxPoint(9,64));
	frameTime = new wxStaticText(renderingPanel, wxID_ANY,wxT("0 ms")  ,wxPoint(80,44), wxSize(90,15), wxALIGN_RIGHT | wxST_NO_AUTORESIZE);
	fps		  = new wxStaticText(renderingPanel, wxID_ANY,wxT("0 fps") ,wxPoint(80,64), wxSize(90,15), wxALIGN_RIGHT | wxST_NO_AUTORESIZE);


	new wxStaticBox (structurePanel, wxID_ANY,wxT("Acceleration structure") ,wxDefaultPosition,wxSize(180,120));
	new wxStaticText(structurePanel, wxID_ANY,wxT("Type"),wxPoint(9,24));
	structureBox = new wxComboBox (structurePanel,Structure_Combo,wxEmptyString,wxPoint(92,20),wxSize(84,21),3,STRUCTURE_NAME,wxCB_READONLY);
	structureBox->SetSelection(0);


	SetSizer(mainSizer);

	settingsPanel->SetMinSize(wxSize(180,512));

	//Resize the main window, so all controls fit inside.
	Layout();
	Fit();

	wxImage::AddHandler(new wxPNGHandler);
}

void MainWindow::onViewMode(wxCommandEvent& event) {
	if(event.GetId() == Tool_Omni) {
		if(omni->IsToggled()) {
			toolBar->ToggleTool(Tool_Walk,false);
			canvas->setViewMode(true);
		} else {
			toolBar->ToggleTool(Tool_Omni,true);
		};
	} else {
		if(walk->IsToggled()) {
			toolBar->ToggleTool(Tool_Omni,false);
			canvas->setViewMode(false);
		} else {
			toolBar->ToggleTool(Tool_Walk,true);
		};
	};
};

void MainWindow::onMainAxis(wxCommandEvent& event) {
	int curId = event.GetId();
	if(curId != axisLast) {
		toolBar->ToggleTool(axisLast,false);
		canvas->setMainAxis(curId - Tool_X);
		axisLast = curId;
	} else {
		toolBar->ToggleTool(curId,true);
	};
};

void MainWindow::setRenderTime(float time) {
	char text[128];

	sprintf(text,"%.2f ms",time);
	frameTime->SetLabel(wxString(text,wxConvISO8859_1));

	time = 1.0f/(time*0.001f);
	sprintf(text,"%.2f fps",time);
	fps->SetLabel(wxString(text,wxConvISO8859_1));

};

void MainWindow::OnQuit(wxCommandEvent& WXUNUSED(event)) {
    Close(true);
}

void MainWindow::OnResolutionChange(wxCommandEvent& event) {
	unsigned resolution = resolutionBox->GetSelection();

	//canvas->SetSize(RESOLUTIONS[resolution].width,RESOLUTIONS[resolution].height);
	canvas->SetMinSize(wxSize(RESOLUTIONS[resolution].width,RESOLUTIONS[resolution].height));
	settingsPanel->SetMinSize(wxSize(180,RESOLUTIONS[resolution].height));

	Layout();
	Fit();

	canvas->setNewResolution(resolution);
};

void MainWindow::OnStructureChange(wxCommandEvent& event) {
	canvas->setNewStructure(structureBox->GetSelection());
};

void MainWindow::OnFileOpen(wxCommandEvent& event) {
	wxFileDialog fileOpenDialog(this);
	fileOpenDialog.SetWildcard(FORMAT_FILTER);
	fileOpenDialog.SetStyle(wxFD_OPEN | wxFD_FILE_MUST_EXIST);
	if(fileOpenDialog.ShowModal() == wxID_OK) {
		char*  outFileName = new char[4096];
		wxString fileName  = fileOpenDialog.GetPath();
		#ifdef wxUSE_UNICODE
			const wchar_t* cFileName = fileName.c_str();
			wxConvLibc.FromWChar(outFileName,4096,cFileName);
		#else 
			const char*    cFileName = fileName.c_str();
			strcpy(outFileName,cFileName);
		#endif
		canvas->loadModel(outFileName);
		delete outFileName;
	};
};

MainWindow::~MainWindow(void) {
}
