#pragma once

#include "Utils/GLEE/glew.h"

#include <wx/wx.h>
#include <wx/glcanvas.h>

#include "Gpu/Wrapper.h"


enum Direction {
	MOVE_LEFT,MOVE_RIGHT,MOVE_FRONT,MOVE_BACK,BANK_LEFT,BANK_RIGHT
};

class TextureDisplay : public wxGLCanvas {
	struct Texture {
		GLuint		texture;
		unsigned	width;
		unsigned	height;
	};

	CudaRenderer*	 renderer;
	Scene			*scene;
	Camera			*camera;

	Texture textureBuild,textureLoad;
	GLuint  pbo;

	bool init;
	bool modelLoaded;
	unsigned splashScreen;

	bool	mouseDown;
	wxCoord	lastX;
	wxCoord	lastY;
	float   clickAngleX;
	float   clickAngleY;

	bool  isOmni;

	unsigned mainAxis;
	float angleX;
	float angleY;
	float angleZ;
	
	float distance;

	bool moveLeft,moveRight,moveFront,moveBack,bankLeft,bankRight;

	unsigned width;
	unsigned height;
	unsigned structure;


	void recomputeCamera();
	void loadTexture(char* filename,Texture& texture);
	void drawCenter(Texture &texture);

public:
	TextureDisplay(	wxWindow *parent, wxWindowID id = wxID_ANY,
					const wxPoint& pos = wxDefaultPosition,
					const wxSize& size = wxDefaultSize, long style = 0,
					const wxString& name = wxT("TestGLCanvas"));

	~TextureDisplay(void);

	void loadModel(const char* filename);
	void setNewResolution(unsigned n);
	void setNewStructure(unsigned newStructure);

	void setViewMode(bool     omni);
	void setMainAxis(unsigned axis);

	void keyDown(Direction key);	
	void keyUp  (Direction key);

	void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
	void OnMouse(wxMouseEvent& event);

	DECLARE_EVENT_TABLE();
};
