/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SceneNode.h"
#include <vector_functions.h>

SceneLoader* SceneLoader::objInstance=0;

SceneLoader::SceneLoader() {
};

SceneLoader::~SceneLoader() {
	//Delete all the file loader objects.
	std::map<std::string,FileLoader*>::iterator cur,end;

	cur = loaderMap.begin();
	end = loaderMap.end();

	while(cur!=end) {
		delete cur->second;
		cur++;
	};
};

unsigned SceneLoader::addLoader(FileLoader* fileLoader,std::string name) {
	loaderMap.insert(std::pair<std::string,FileLoader*>(name,fileLoader));
	return 0;
};

Scene*	SceneLoader::loadScene(const char* filename,bool recenter) {
	//Prepare material manager for use
	std::vector<Material> &materials = MaterialManager::instance()->getMaterials();

	Object*	object = load(filename);	//Load object in file

	//If the file format doesn't provide materials, insert at least one default material.
	if(materials.size() == 0) {
		materials.push_back(Material());
		materials.back().ambient  = make_float3(0.4f,0.4f,0.4f);
		materials.back().diffuse  = make_float3(0.8f,0.8f,0.8f);
		materials.back().specular = make_float3(0.3f,0.3f,0.3f);
		materials.back().shinines = 5.0f;
		materials.back().reflectivity = 0.0f;
	};

	float3 center;
	center.x = (object->maximum.x + object->minimum.x)*0.5f;
	center.y = (object->maximum.y + object->minimum.y)*0.5f;
	center.z = (object->maximum.z + object->minimum.z)*0.5f;

	Transformation* transformation = new Transformation();
	transformation->IdentityMatrix();

	if(recenter)
		transformation->translate(-center.x,-center.y,-center.z);

	Scene* scene = new Scene;
	scene->push_back(ObjectInstance(transformation,object));

	return scene;
};

Object*	 SceneLoader::load(const char* filename) {
	//Find the file extension and quit if it has none.
	const char* extension = strrchr(filename,'.');
	if(!extension || (*extension==0) ) return 0;
	extension++;

	//Try to find the loader object for the given file extension and use it to load the file.
	std::map<std::string,FileLoader*>::iterator result = loaderMap.find(extension);
	if(result!=loaderMap.end()) {
		return result->second->loadFile(filename);	
	} else
		return 0;
};