/*
**	  GPU Raytrace - Diploma thesis, Czech technical university
**	  Copyright (C) 2008 Martin Zlatuska
**
**	  This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#include <xutility>

//! Plain C array iterator.
/**
*	This class provide iterator mnemonics for plain C array. This was mainly implemented so the
*	more efficient std::sort can by used for plain array instead of library qsort function.
*/
template<typename T>
class PlainIterator {
	T*	element;

public:
	typedef std::random_access_iterator_tag iterator_category;
	typedef T								value_type;
	typedef size_t							difference_type;
	typedef T*								pointer;
	typedef T&								reference;


	PlainIterator(T* element):element(element) {};
	PlainIterator();

	reference operator*() {
		return *element;
	};

	pointer   operator->() {
		return element;
	};

	PlainIterator& operator++() {
		element++;
		return *this;
	};
	
	PlainIterator operator++(int) {
		PlainIterator original(element);
		element++;
		return original;
	};

	PlainIterator& operator--() {
		element--;
		return *this;
	};

	PlainIterator  operator--(int) {
		PlainIterator original(element);
		element--;
		return original;
	};
	

	PlainIterator& operator+=(difference_type dist) {
		element+=dist;
		return *this;
	};

	PlainIterator  operator+(difference_type dist) {
		return PlainIterator(element + dist);
	};

	PlainIterator& operator-=(difference_type dist) {
		element-=dist;
		return *this;
	};
	
	PlainIterator  operator-(difference_type dist) {
		return PlainIterator(element - dist);
	};

	difference_type operator-(const PlainIterator& right) {
		return element - right.element;
	};

	reference operator[](difference_type off) {
		return element[off];
	};

	bool operator==(const PlainIterator& right) {
		return element == right.element;
	};
	
	bool operator!=(const PlainIterator& right) {
		return element != right.element;
	};

	bool operator<(const PlainIterator& right) {
		return element < right.element;
	};
};